"""Support sorting for SQLAlchemy ORM."""
from __future__ import annotations

from typing import TYPE_CHECKING, Union, cast

from sqlalchemy import Column

from muffin_rest.sorting import Sort, Sorting

if TYPE_CHECKING:
    from .types import TVCollection


class SASort(Sort):
    """Sorter for Peewee."""

    async def apply(
        self, collection: TVCollection, *, desc: bool = False, **_,
    ) -> TVCollection:
        """Sort the collection."""
        field = self.field
        if desc and isinstance(field, Column):
            field = field.desc()

        return collection.order_by(field)


class SASorting(Sorting):
    """Sort Peewee ORM Queries."""

    MUTATE_CLASS = SASort

    def convert(self, obj: Union[str, Column, SASort], **meta):
        """Prepare sorters."""
        from . import SARESTHandler

        if isinstance(obj, SASort):
            return obj

        handler = cast(SARESTHandler, self.handler)

        if isinstance(obj, Column):
            name, field = obj.name, obj

        else:
            name = obj
            field = meta.get("field", handler.meta.table.c.get(name))

        if field is not None:
            sort = self.MUTATE_CLASS(name, field=field, **meta)
            if sort.meta.get("default"):
                self.default.append(sort)

            return sort

    def sort_default(self, collection: TVCollection) -> TVCollection:
        """Sort collection by default."""
        return collection.order_by(
            *[
                sort.field.desc() if sort.meta["default"] == "desc" else sort.field
                for sort in self.default
            ],
        )



"""Support for Peewee ORM (https://github.com/coleifer/peewee)."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Tuple, Type, Union, cast, overload

import marshmallow as ma
import peewee as pw
from apispec.ext.marshmallow import MarshmallowPlugin
from marshmallow_peewee import ForeignKey
from peewee_aio.model import AIOModel, AIOModelSelect

from muffin_rest.errors import APIError
from muffin_rest.handler import RESTBase
from muffin_rest.peewee.openapi import PeeweeOpenAPIMixin

from .options import PWRESTOptions
from .schemas import EnumField
from .types import TVModel

if TYPE_CHECKING:
    from muffin import Request
    from peewee_aio.types import TVAIOModel


# XXX: Patch apispec.MarshmallowPlugin to support ForeignKeyField
MarshmallowPlugin.Converter.field_mapping[ForeignKey] = ("integer", None)

assert issubclass(EnumField, ma.fields.Field)  # just register EnumField


class PWRESTBase(RESTBase[TVModel], PeeweeOpenAPIMixin):
    """Support Peeweee."""

    if TYPE_CHECKING:
        resource: TVModel
        collection: Union[AIOModelSelect, pw.ModelSelect]

    meta: PWRESTOptions
    meta_class: Type[PWRESTOptions] = PWRESTOptions

    @overload
    async def prepare_collection(
        self: PWRESTBase[TVAIOModel],
        _: Request,
    ) -> AIOModelSelect[TVAIOModel]:
        ...

    @overload
    async def prepare_collection(
        self: PWRESTBase[pw.Model],
        _: Request,
    ) -> pw.ModelSelect:
        ...

    # NOTE: there is not a default sorting for peewee (conflict with muffin-admin)
    async def prepare_collection(self, _: Request):
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return self.meta.model.select()

    async def prepare_resource(self, request: Request) -> Optional[TVModel]:
        """Load a resource."""
        pk = request["path_params"].get(self.meta.name_id)
        if not pk:
            return None

        meta = self.meta

        try:
            resource = await meta.manager.fetchone(
                self.collection.where(meta.model_pk == pk),
            )
        except Exception:  # noqa:
            resource = None

        if resource is None:
            raise APIError.NOT_FOUND("Resource not found")

        return resource

    @overload
    async def paginate(
        self: PWRESTBase[TVAIOModel], _: Request, *, limit: int = 0, offset: int = 0
    ) -> Tuple[AIOModelSelect[TVAIOModel], int]:
        ...

    @overload
    async def paginate(
        self: PWRESTBase[pw.Model], _: Request, *, limit: int = 0, offset: int = 0
    ) -> Tuple[pw.ModelSelect, int]:
        ...

    async def paginate(self, _: Request, *, limit: int = 0, offset: int = 0):
        """Paginate the collection."""
        cqs = cast(pw.ModelSelect, self.collection.order_by())
        if cqs._group_by:  # type: ignore[misc]
            cqs._returning = cqs._group_by  # type: ignore[misc]
            cqs._having = None  # type: ignore[misc]
        count = await self.meta.manager.count(cqs)

        return (
            cast(pw.ModelSelect, self.collection.offset(offset).limit(limit)),  # type: ignore[misc]
            count,
        )

    async def get(self, request, *, resource: Optional[TVModel] = None) -> Any:
        """Get resource or collection of resources."""
        if resource:
            return await self.dump(request, resource)

        resources = await self.meta.manager.fetchall(self.collection)
        return await self.dump(request, resources, many=True)

    async def save(self, request: Request, resource: TVModel, *, update=False):
        """Save the given resource."""
        meta = self.meta
        manager = meta.manager
        if issubclass(meta.model, AIOModel):
            await resource.save()
        else:
            await manager.save(resource)

        return resource

    async def remove(self, request: Request, resource: Optional[TVModel] = None):
        """Remove the given resource."""
        meta = self.meta
        if resource:
            resources = [resource]

        else:
            data = await request.data()
            if not data:
                return

            model_pk = cast(pw.Field, meta.model_pk)
            resources = await meta.manager.fetchall(
                self.collection.where(model_pk << data),
            )

        if not resources:
            raise APIError.NOT_FOUND()

        is_aiomodel = issubclass(meta.model, AIOModel)
        if is_aiomodel:
            for res in resources:
                await res.delete_instance(recursive=meta.delete_recursive)
        else:
            for res in resources:
                await meta.manager.delete_instance(res, recursive=meta.delete_recursive)

    async def delete(self, request: Request, resource: Optional[TVModel] = None):
        return await self.remove(request, resource)

    async def get_schema(
        self, request: Request, resource: Optional[TVModel] = None, **_
    ) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get("schema_only"),
            exclude=request.url.query.get("schema_exclude", ()),
        )


class PWRESTHandler(PWRESTBase[TVModel], PeeweeOpenAPIMixin):
    meta: PWRESTOptions



"""Implement a base class for API."""

from __future__ import annotations

import dataclasses as dc
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union, overload

from http_router import Router
from muffin.utils import TV, to_awaitable

from .errors import InvalidEnpointError
from .openapi import render_openapi

if TYPE_CHECKING:
    import muffin

    from muffin_rest.types import TAuth, TVAuth, TVHandler

REDOC_TEMPLATE = Path(__file__).parent.joinpath("redoc.html").read_text()
SWAGGER_TEMPLATE = Path(__file__).parent.joinpath("swagger.html").read_text()


@dc.dataclass
class API:
    """Initialize an API."""

    def __init__(
        self,
        app: Optional[muffin.Application] = None,
        prefix: str = "",
        *,
        openapi: bool = True,
        servers: Optional[List] = None,
        **openapi_info,
    ):
        """Post initialize the API if we have an application already."""
        self.app = app
        self.prefix = prefix

        self.openapi = openapi
        self.openapi_options: Dict[str, Any] = {"info": openapi_info}
        if servers:
            self.openapi_options["servers"] = servers

        self.authorize: TAuth = to_awaitable(lambda _: True)
        self.router = Router()

        if app:
            self.setup(app, prefix=prefix)

    def __repr__(self):
        """Stringify the API."""
        return f"<API { self.prefix }>"

    @property
    def logger(self):
        """Proxy the application's logger."""
        if self.app is None:
            raise RuntimeError("API is not initialized yet")

        return self.app.logger

    def setup(
        self,
        app: muffin.Application,
        *,
        prefix: str = "",
        openapi: Optional[bool] = None,
        servers: Optional[List] = None,
        **openapi_info,
    ):
        """Initialize the API."""
        self.app = app
        self.prefix = (prefix or self.prefix).rstrip("/")

        # Setup routing
        self.router.trim_last_slash = self.app.router.trim_last_slash
        self.router.validator = self.app.router.validator
        self.app.router.route(self.prefix)(self.router)

        if openapi is not None:
            self.openapi = openapi

        if openapi_info:
            self.openapi_options["info"] = openapi_info

        if servers:
            self.openapi_options["servers"] = servers

        # Setup API Docs
        if not self.openapi:
            return

        async def openapi_json(request):
            return render_openapi(self, request=request)

        self.router.route("/swagger")(swagger)
        self.router.route("/redoc")(redoc)
        self.router.route("/openapi.json")(openapi_json)

    @overload
    def route(self, obj: str, *paths: str, **params) -> Callable[[TV], TV]:
        ...

    @overload
    def route(self, obj: TVHandler, *paths: str, **params) -> TVHandler:
        ...

    def route(
        self, obj: Union[str, TVHandler], *paths: str, **params
    ) -> Union[Callable[[TV], TV], TVHandler]:
        """Route an endpoint by the API."""
        from .handler import RESTBase

        def wrapper(cb):
            cb._api = self
            return self.router.route(*paths, **params)(cb)

        if isinstance(obj, str):
            paths = (obj, *paths)
            return wrapper

        # Generate URL paths automatically
        if issubclass(obj, RESTBase):
            obj._api = self
            return self.router.route(*paths, **params)(obj)

        raise InvalidEnpointError

    def authorization(self, auth: TVAuth) -> TVAuth:
        """Bind an authorization flow to self API."""
        self.authorize = auth
        return auth


async def swagger(_) -> str:
    """Get the Swagger UI."""
    return SWAGGER_TEMPLATE


async def redoc(_) -> str:
    """Get the ReDoc UI."""
    return REDOC_TEMPLATE



"""Create openapi schema from the given API."""
from __future__ import annotations

import inspect
import re
from contextlib import suppress
from functools import partial
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, cast

from apispec import utils
from apispec.core import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin
from asgi_tools.response import CAST_RESPONSE
from asgi_tools.types import TJSON
from http_router.routes import DynamicRoute, Route
from muffin import Response

from . import LIMIT_PARAM, OFFSET_PARAM

if TYPE_CHECKING:
    from .options import RESTOptions

with suppress(ImportError):
    from apispec import yaml_utils

locals().setdefault("yaml_utils", None)


DEFAULT_METHODS = ("get",)
HTTP_METHODS = [
    "GET",
    "POST",
    "PUT",
    "PATH",
    "DELETE",
    "HEAD",
    "OPTIONS",
    "TRACE",
    "CONNECT",
]
RE_URL = re.compile(r"<(?:[^:<>]+:)?([^<>]+)>")  # TODO: Not used
SKIP_PATH = {"/openapi.json", "/swagger", "/redoc"}


def render_openapi(api, request=None):
    """Prepare openapi specs."""
    # Setup Specs
    options = dict(api.openapi_options)
    if request:
        options.setdefault(
            "servers",
            [{"url": str(request.url.with_query("").with_path(api.prefix))}],
        )

    spec = APISpec(
        options["info"].pop("title", f"{ api.app.cfg.name.title() } API"),
        options["info"].pop("version", "1.0.0"),
        options.pop("openapi_version", "3.0.3"),
        **options,
        plugins=[MarshmallowPlugin()],
    )
    tags = {}

    # Setup Authorization
    if api.authorize:
        _, _, schema = parse_docs(api.authorize)
        spec.options["security"] = []
        for key, value in schema.items():
            spec.components.security_scheme(key, value)
            spec.options["security"].append({key: []})

    # Setup Paths
    routes = api.router.routes()
    for route in routes:
        if route.path in SKIP_PATH:
            continue

        spec.path(route.path, **route_to_spec(route, spec, tags))

    return spec.to_dict()


def route_to_spec(route: Route, spec: APISpec, tags: Dict) -> Dict:
    """Convert the given router to openapi operations."""
    results: Dict[str, Any] = {"parameters": [], "operations": {}}
    if isinstance(route, DynamicRoute):
        for param in route.params:
            results["parameters"].append({"in": "path", "name": param})

    target = cast(Callable, route.target)
    if isinstance(target, partial):
        target = target.func

    if hasattr(target, "openapi"):
        results["operations"] = target.openapi(route, spec, tags)
        return results

    summary, desc, schema = parse_docs(target)
    responses = return_type_to_response(target)
    for method in route_to_methods(route):
        results["operations"][method] = {
            "summary": summary,
            "description": desc,
            "responses": responses,
        }

    results["operations"] = merge_dicts(results["operations"], schema)
    return results


def parse_docs(cb: Callable) -> Tuple[str, str, Dict]:
    """Parse docs from the given callback."""
    if yaml_utils is None:
        return "", "", {}

    docs = cb.__doc__ or ""
    schema = yaml_utils.load_yaml_from_docstring(docs)
    docs = docs.split("---")[0]
    docs = utils.dedent(utils.trim_docstring(docs))
    summary, _, description = docs.partition("\n\n")
    return summary, description.strip(), schema


def merge_dicts(source: Dict, merge: Dict) -> Dict:
    """Merge dicts."""
    return dict(
        source,
        **{
            key: (
                (
                    merge_dicts(source[key], merge[key])
                    if isinstance(source[key], dict) and isinstance(merge[key], dict)
                    else (
                        source[key] + merge[key]
                        if isinstance(source[key], list) and isinstance(merge[key], list)
                        else merge[key]
                    )
                )
                if key in source
                else merge[key]
            )
            for key in merge
        },
    )


def route_to_methods(route: Route) -> List[str]:
    """Get sorted methods from the route."""
    methods = [m for m in HTTP_METHODS if m in (route.methods or [])]
    return [m.lower() for m in methods or DEFAULT_METHODS]


def return_type_to_response(fn: Callable) -> Dict:
    """Generate reponses specs based on the given function's return type."""
    responses: Dict[int, Dict] = {}
    return_type = fn.__annotations__.get("return")
    if return_type is None:
        return responses

    return_type = CAST_RESPONSE.get(return_type, return_type)
    if return_type is None:
        return responses

    if (
        inspect.isclass(return_type)
        and issubclass(return_type, Response)
        and return_type.content_type
    ):
        responses[return_type.status_code] = {
            "description": HTTPStatus(return_type.status_code).description,
            "content": {return_type.content_type: {}},
        }
    return responses


class OpenAPIMixin:
    """Render an endpoint to openapi specs."""

    meta: RESTOptions

    @classmethod
    def openapi(cls, route: Route, spec: APISpec, tags: Dict) -> Dict:  # noqa: C901
        """Get openapi specs for the endpoint."""
        meta = cls.meta
        if getattr(meta, meta.base_property, None) is None:
            return {}

        operations: Dict = {}
        summary, desc, schema = parse_docs(cls)
        if cls not in tags:
            tags[cls] = meta.name
            spec.tag({"name": meta.name, "description": summary})
            schema_cls = meta.Schema
            if schema_cls.__name__ not in spec.components.schemas:
                spec.components.schema(meta.Schema.__name__, schema=meta.Schema)

        schema_ref = {"$ref": f"#/components/schemas/{ meta.Schema.__name__ }"}
        for method in route_to_methods(route):
            operations[method] = {"tags": [tags[cls]]}
            is_resource_route = isinstance(route, DynamicRoute) and route.params.get(
                meta.name_id,
            )

            if method == "get" and not is_resource_route:
                operations[method]["parameters"] = []
                if meta.sorting:
                    operations[method]["parameters"].append(meta.sorting.openapi)

                if meta.filters:
                    operations[method]["parameters"].append(meta.filters.openapi)

                if meta.limit:
                    operations[method]["parameters"].append(
                        {
                            "name": LIMIT_PARAM,
                            "in": "query",
                            "schema": {
                                "type": "integer",
                                "minimum": 1,
                                "maximum": meta.limit,
                            },
                            "description": "The number of items to return",
                        },
                    )
                    operations[method]["parameters"].append(
                        {
                            "name": OFFSET_PARAM,
                            "in": "query",
                            "schema": {"type": "integer", "minimum": 0},
                            "description": "The offset of items to return",
                        },
                    )

            # Update from the method
            meth = getattr(cls, method, None)
            if isinstance(route.target, partial) and "__meth__" in route.target.keywords:
                meth = getattr(cls, route.target.keywords["__meth__"], None)

            elif method in {"post", "put"}:
                operations[method]["requestBody"] = {
                    "required": True,
                    "content": {"application/json": {"schema": schema_ref}},
                }

            if meth:
                (operations[method]["summary"], operations[method]["description"], mschema) = (
                    parse_docs(meth)
                )
                return_type = meth.__annotations__.get("return")
                if return_type in ("JSONType", TJSON):
                    responses = {
                        200: {
                            "description": "Request is successfull",
                            "content": {"application/json": {"schema": schema_ref}},
                        },
                    }
                else:
                    responses = return_type_to_response(meth)

                operations[method]["responses"] = responses
                operations[method] = merge_dicts(operations[method], mschema)

        return merge_dicts(operations, schema)



from typing import Any, Dict, TypeVar

from sqlalchemy import sql

TVCollection = TypeVar("TVCollection", bound=sql.Select)

TResource = Dict[str, Any]
TVResource = TypeVar("TVResource", bound=TResource)



"""Mongo DB support."""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple, Type, cast

import bson
from bson.errors import InvalidId

from muffin_rest.errors import APIError
from muffin_rest.handler import RESTHandler, RESTOptions
from muffin_rest.mongo.filters import MongoFilters
from muffin_rest.mongo.schema import MongoSchema
from muffin_rest.mongo.sorting import MongoSorting
from muffin_rest.mongo.utils import MongoChain

from .types import TVResource

if TYPE_CHECKING:
    import marshmallow as ma
    from motor import motor_asyncio as motor
    from muffin import Request


class MongoRESTOptions(RESTOptions):
    """Support Mongo DB."""

    filters_cls: Type[MongoFilters] = MongoFilters
    sorting_cls: Type[MongoSorting] = MongoSorting
    schema_base: Type[MongoSchema] = MongoSchema

    aggregate: Optional[List] = None  # Support aggregation. Set to pipeline.
    collection_id: str = "_id"
    collection: motor.AsyncIOMotorCollection

    base_property: str = "collection"

    Schema: Type[MongoSchema]

    def setup(self, cls):
        """Prepare meta options."""
        if self.collection is None:
            raise RuntimeError("MongoResthandler.Meta.collection is required")
        self.name = getattr(self, "name", self.collection.name.lower())
        self.name = self.name or self.collection.name.lower()
        super().setup(cls)


class MongoRESTHandler(RESTHandler[TVResource]):
    """Support Mongo DB."""

    meta: MongoRESTOptions
    meta_class: Type[MongoRESTOptions] = MongoRESTOptions

    async def prepare_collection(self, _: Request) -> MongoChain:
        """Initialize Peeewee QuerySet for a binded to the resource model."""
        return MongoChain(self.meta.collection)

    async def paginate(
        self, _: Request, *, limit: int = 0, offset: int = 0
    ) -> Tuple[motor.AsyncIOMotorCursor, int]:
        """Paginate collection."""
        if self.meta.aggregate:
            pipeline_all = [*self.meta.aggregate, {"$skip": offset}, {"$limit": limit}]
            pipeline_num = [
                *self.meta.aggregate,
                {"$group": {self.meta.collection_id: None, "total": {"$sum": 1}}},
            ]
            counts = list(self.collection.aggregate(pipeline_num))
            return (
                self.collection.aggregate(pipeline_all),
                counts and counts[0]["total"] or 0,  # type: ignore[]
            )
        total = await self.collection.count()
        return self.collection.skip(offset).limit(limit), total

    async def get(self, request, *, resource: Optional[TVResource] = None):
        """Get resource or collection of resources."""
        if resource:
            return await self.dump(request, resource)

        docs = await self.collection.to_list(None)
        return await self.dump(request, docs, many=True)

    async def prepare_resource(self, request: Request) -> Optional[TVResource]:
        """Load a resource."""
        pk = request["path_params"].get(self.meta.name_id)
        if not pk:
            return None

        try:
            return await self.collection.find_one(
                {self.meta.collection_id: bson.ObjectId(pk)},
            )
        except InvalidId as exc:
            raise APIError.NOT_FOUND() from exc

    async def get_schema(
        self, request: Request, resource: Optional[TVResource] = None, **_
    ) -> ma.Schema:
        """Initialize marshmallow schema for serialization/deserialization."""
        return self.meta.Schema(
            instance=resource,
            only=request.url.query.get("schema_only"),
            exclude=request.url.query.get("schema_exclude", ()),
        )

    async def save(
        self, _: Request, resource: TVResource, *, update=False
    ) -> TVResource:
        """Save the given resource."""
        meta = self.meta
        collection_id = meta.collection_id
        if update:
            await self.collection.replace_one(
                {collection_id: resource[collection_id]}, resource
            )

        else:
            updated = await meta.collection.insert_one(resource)
            resource[collection_id] = updated.inserted_id

        return resource

    async def delete(self, request: Request, resource: Optional[TVResource] = None):
        """Remove the given resource(s)."""
        meta = self.meta
        oids = (
            [resource[meta.collection_id]]
            if resource
            else cast(List[str], await request.data())
        )
        if not oids:
            raise APIError.NOT_FOUND()

        if not isinstance(oids, list):
            raise APIError.BAD_REQUEST()

        oids = [bson.ObjectId(idx) for idx in oids]
        await meta.collection.delete_many({meta.collection_id: {"$in": oids}})



"""Support API filters."""
from __future__ import annotations

import operator
from typing import TYPE_CHECKING, Any, Callable, Dict, Mapping, Optional, Tuple  # py38

import marshmallow as ma
from asgi_tools._compat import json_loads  # type: ignore[]

from .utils import Mutate, Mutator

if TYPE_CHECKING:
    from muffin import Request

    from .types import TVCollection

FILTERS_PARAM = "where"


class Filter(Mutate):

    """Base filter class."""

    operators: Dict[str, Callable] = {
        "$lt": operator.lt,
        "$le": operator.le,
        "$gt": operator.gt,
        "$ge": operator.ge,
        "$eq": operator.eq,
        "$ne": operator.ne,
        "$in": operator.contains,
        "$nin": lambda v, c: v not in c,
    }
    operators["<"] = operators["$lt"]
    operators["<="] = operators["$le"]
    operators[">"] = operators["$gt"]
    operators[">="] = operators["$ge"]
    operators["=="] = operators["$eq"]
    operators["!="] = operators["$ne"]
    operators["<<"] = operators["$in"]

    list_ops = ["$in", "<<"]

    field: Any = None
    schema_field: ma.fields.Field = ma.fields.Raw()
    default_operator = "$eq"

    def __init__(
        self,
        name: str,
        *,
        field: Any = None,
        schema_field: Optional[ma.fields.Field] = None,
        operator: Optional[str] = None,
        **meta,
    ):
        """Initialize filter.

        :param name: The filter's name
        :param attr: Column/Property name.
        :param fields: Marshmallow Field instance

        """
        super(Filter, self).__init__(name, **meta)
        self.field = field if field is not None else self.field
        self.schema_field: ma.fields.Field = schema_field or self.schema_field
        self.default_operator = operator or self.default_operator

    async def apply(self, collection: Any, data: Optional[Mapping] = None):
        """Filter given collection."""
        if not data:
            return None, collection

        try:
            ops = self.parse(data)
        except ma.ValidationError:
            return None, collection

        if ops:
            collection = await self.filter(collection, *ops)

        return ops, collection

    async def filter(self, collection, *ops: Tuple[Callable, Any], **_):
        """Apply the filter to collection."""

        def validator(obj):
            return all(op(get_value(obj, self.name), val) for op, val in ops)

        return [o for o in collection if validator(o)]

    def parse(self, data: Mapping) -> Tuple[Tuple[Callable, Any], ...]:
        """Parse operator and value from filter's data."""
        val = data.get(self.name, ma.missing)
        if not isinstance(val, dict):
            return (
                (
                    self.operators[self.default_operator],
                    self.schema_field.deserialize(val),
                ),
            )

        return tuple(
            (
                self.operators[op],
                (
                    (self.schema_field.deserialize(val))
                    if op not in self.list_ops
                    else [self.schema_field.deserialize(v) for v in val]
                ),
            )
            for (op, val) in val.items()
            if op in self.operators
        )


class Filters(Mutator):

    """Build filters for handlers."""

    MUTATE_CLASS = Filter
    mutations: Mapping[str, Filter]

    async def apply(
        self, request: Request, collection: TVCollection
    ) -> Tuple[TVCollection, Dict[str, Any]]:
        """Filter the given collection."""
        raw_data = request.url.query.get(FILTERS_PARAM)
        filters = {}
        if raw_data is not None:
            try:
                data = json_loads(raw_data)
                assert isinstance(data, dict)
                mutations = self.mutations
                for name in data:
                    if name in mutations:
                        ops, collection = await mutations[name].apply(collection, data)
                        filters[name] = ops

            except (ValueError, TypeError, AssertionError):
                api = self.handler._api
                api.logger.warning("Invalid filters data: %s", request.url)

        return collection, filters

    def convert(self, obj, **meta):
        """Convert params to filters."""
        if isinstance(obj, self.MUTATE_CLASS):
            return obj

        field = meta.pop("field", None) or obj
        schema_field = meta.pop("schema_field", None)
        if schema_field is None and field:
            schema_field = self.handler.meta.Schema._declared_fields.get(field)
        return self.MUTATE_CLASS(obj, field=field, schema_field=schema_field, **meta)

    @property
    def openapi(self) -> Dict:
        """Prepare OpenAPI params."""
        return {
            "name": FILTERS_PARAM,
            "in": "query",
            "description": str(self),
            "content": {"application/json": {"schema": {"type": "object"}}},
        }


def get_value(obj, name: str):
    """Get value from object by name."""
    if isinstance(obj, dict):
        return obj.get(name)

    return getattr(obj, name, obj)



import marshmallow as ma
from marshmallow_peewee import DefaultConverter
from muffin_peewee.fields import IntEnumField, StrEnumField, URLField


class EnumField(ma.fields.Field):
    default_error_messages = {
        "unknown": "Must be one of: {choices}.",
    }

    def __init__(self, enum, **kwargs):
        self.enum = enum
        self.choices_text = ", ".join([str(c.value) for c in enum])
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None

        try:
            return value.value
        except AttributeError:
            raise ma.ValidationError(f"{obj}: {attr} value is invalid: {value}") from None

    def _deserialize(self, value, attr, data, **kwargs):
        try:
            return self.enum(value)
        except ValueError as error:
            raise self.make_error("unknown", choices=self.choices_text) from error


@DefaultConverter.register(StrEnumField)
@DefaultConverter.register(IntEnumField)
def build_field(field, opts, **params):
    params.pop("validate", None)
    return EnumField(field.enum, **params)


DefaultConverter.register(URLField, ma.fields.Url)

# ruff: noqa: ARG001, ARG002



"""Serialize/deserialize results from mongo db."""

import bson
import marshmallow as ma


class ObjectId(ma.fields.Field):

    """ObjectID Marshmallow Field."""

    def _deserialize(self, value, _, __):
        try:
            return bson.ObjectId(value)
        except ValueError as exc:
            raise ma.ValidationError("invalid ObjectId `%s`" % value) from exc

    def _serialize(self, value, _, __):
        if value is None:
            return ma.missing
        return str(value)


class MongoSchema(ma.Schema):

    """Serialize/deserialize results from mongo."""

    _id = ObjectId()

    def __init__(self, instance=None, **kwargs):
        """Initialize the schema."""
        self.instance = instance
        super(MongoSchema, self).__init__(**kwargs)

    @ma.post_load
    def make_instance(self, data, **__):
        """Build object from data."""
        if self.instance is not None:
            self.instance.update(data)
            return self.instance

        return data

    def load(self, data, instance=None, *args, **kwargs):
        """Load data."""
        self.instance = instance or self.instance
        return super(MongoSchema, self).load(data, *args, **kwargs)
